"""Runs the debug server on the dersire file

Usage:
    vizual [--file=FILE]
    
"""

import docopt
import os

def main():
    args = docopt.docopt(__doc__)
    target = args['--file'] # something.py filepath relative to the terminal.
    server = os.path.join(os.path.dirname(__file__), 'server.py')

    cmd = "(trap 'kill 0' SIGINT; python3 " + target + " & python3 " + server + ")"

    os.system(cmd)

if __name__ == '__main__':
    main()


from time import time, sleep
import pandas as pd
import requests

def table_head(df):
        return df.head()

class Vizual:
    def __init__(self, server_ip='http://0.0.0.0:8080', live_post=False):
        sleep(1)
        self._ip = server_ip
        self._channel_context = ''
        self._context_stack = []
        self._is_test = True
        self.message_q = []
        self.live_post = live_post

    def post(self, url, msg=None):
        if self.live_post:
            requests.post(self._ip + url, json=msg)
        else:
            self.message_q.append([url, msg])

    def _post_messages(self):
        while len(self.message_q) > 0:
            url, msg = self.message_q.pop(0)
            requests.post(self._ip + url, json=msg)

    def entry_point(self):
        def entry_wrapper(entry_func):
            def modified_func(*args, **kwargs):
                # Run Test Mode
                entry_func(*args, **kwargs)
                self._is_test = False

                # Run Dev Mode
                res = entry_func(*args, **kwargs)
                
                # Transmit all messages once programme has ended
                self._post_messages()
                return res
            return modified_func
        return entry_wrapper

    def _apply_decorators(self, f, d_fns):
        f_prime = f
        
        for i in range(len(d_fns) - 1,-1,-1):
            decorator = d_fns[i]
            f_prime = decorator(f_prime)

        return f_prime

    def decorate(self, test = [], dev = [], universal = []):
        def decorate_inner(f):
            _test_f = self._apply_decorators(f, test)
            _dev_f = self._apply_decorators(f, dev)

            def modified_function(*args, **kwargs):
                if self._is_test:
                    return _test_f(*args, **kwargs)
                else:
                    return _dev_f(*args, **kwargs)

            return self._apply_decorators(modified_function, universal)

        return decorate_inner

    def _register_channel(self, channel):
        self.post('/channels/%s' % channel, None)

    def _send_debug_message(self, channel, msg):
        self.post('/channels/%s' % channel, msg)

    def _set_channel_context(self, channel):
        self._context_stack.append(channel)

    def _get_channel_context(self):
        return self._context_stack[-1]

    def _pop_channel_context(self):
        self._context_stack.pop(-1)

    def label(self, channel):
        self._register_channel(channel)

        def label_decorator(func):
            def modified_func(*args, **kwargs):
                self._set_channel_context(channel)
                evaluation = func(*args, **kwargs)
                self._pop_channel_context()

                return evaluation
            return modified_func
        return label_decorator

    def table(self, table_name, color='#fff', summary_fn = table_head):
        def table_decorator(func):
            def wrapper(*args, **kwargs):
                table_eval = func(*args, **kwargs)
                
                msg = {
                    'timestamp': time(),
                    'table': summary_fn(table_eval).to_html(border=0),
                    'table_name': table_name,
                    'format': {'color': color}
                }

                self._send_debug_message(self._get_channel_context(), msg)
                
                return table_eval
            return wrapper
        return table_decorator

    def debug(self, template, color='#fff', display_output=True): 
        def debug_decorator(func):
            def wrapper(*args, **kwargs):
                # Frist set the context
                evaluation = func(*args, **kwargs)
                content = template % evaluation if display_output else template

                msg = {
                    'timestamp': time(),
                    'message': content,
                    'format': {'color': color}
                }

                self._send_debug_message(self._get_channel_context(), msg)
                
                return evaluation
            
            return wrapper

        return debug_decorator

    def task(self, task_name, total_iters): 
        def task_decorator(func):
            def wrapper(*args, **kwargs):
                msg = {
                    'task_name': task_name,
                    'total_iters': total_iters,
                }

                self.post('/task/new', msg)

                # Pass thru the intercepted value
                return func(*args, **kwargs)
            
            return wrapper

        return task_decorator

    def task_progress(self, increment): 
        def task_progress_decorator(func):
            def wrapper(*args, **kwargs):
                msg = {
                    'increment': increment,
                }

                self.post('/task/update', msg)

                # Pass thru the intercepted value
                return func(*args, **kwargs)
            
            return wrapper

        return task_progress_decorator

    def ping(self, color='rgb(211,183,124)'):
        # Note pings must be run inside a @debug scope
        # Pings must wrap functions which return boolean expr.
        def ping_decorator(func):
            def wrapper(*args, **kwargs):
                criteria_met = func(*args, **kwargs)

                if criteria_met:
                    msg = {
                        'timestamp': time(),
                        'message': 'Ping!',
                        'format': {'color': color}
                    }
                    
                    self.post('/channels/%s' % self._get_channel_context(), msg)

                return criteria_met
            return wrapper
        return ping_decorator

    def unit_test(self, unit_test_fns):
        # unit_test_fns: list of functions which accept the decorated function
        # as the only argument and returns a boolean true = passed

        def unit_test_decorator(func):
            tests = [] # Trick to ensure we only run the tests once.
            # Only wrapper gets called on each call of func()

            def wrapper(*args, **kwargs):
                if len(tests) == 0:
                    for test_fn in unit_test_fns:
                        msg = {
                            'function': func.__name__,
                            'test_name': test_fn.__name__,
                            'is_passing': test_fn(func)
                        }

                        tests.append(msg)

                    self.post('/tests/%s' % self._get_channel_context(), tests)

                return func(*args, **kwargs)

            return wrapper
        return unit_test_decorator

    def time(self):
        def time_decorator(func):
            def wrapper(*args, **kwargs):
                channel = self._get_channel_context()
                
                a = time()
                eval = func(*args, **kwargs)
                b = time()

                msg = {
                    'function': func.__name__,
                    'duration': b - a,
                }
                    
                self.post('/timing/%s' % channel, msg)

                return eval
            return wrapper
        return time_decorator


from setuptools import setup, find_packages

# Setting up
setup(
      # the name must match the folder name 'verysimplemodule'
      name="vizual",
      version="3.0.0",
      author="Hugo Dolan",
      author_email="<hugojdolan@gmail.com>",
      description="Web UI and decorator synatx for print statement debugging",
      long_description="See readme file",
      packages=['vizual'],
      entry_points={
            'console_scripts': [
                  'vizual = vizual.vizual:main',
            ],
      },
      # This needs to be changed on each new release
      url = 'https://github.com/hugo1005/vizual',
      download_url = 'https://github.com/hugo1005/vizual/archive/refs/tags/v3.tar.gz',
      include_package_data=True,
      install_requires=['Flask','requests','docopt','pandas','numpy'], # add any additional packages that
      # needs to be installed along with your package. Eg: 'caer'
      
      keywords=['python', 'debugging', 'webapp', 'visual'],
      classifiers= [
                    "Development Status :: 5 - Production/Stable",
                    "Intended Audience :: Developers",
                    "Programming Language :: Python :: 2",
                    "Programming Language :: Python :: 3",
                    "Operating System :: MacOS :: MacOS X",
                    "Operating System :: Microsoft :: Windows",
                    ]
      )



import json
from flask import Flask, jsonify, request
import logging

# Server Side
debug_log = {}
timing_log = {}
tests_log = {}
MAIN_TASK = {'task_name':'Main', 'progress':0, 'total_iters': 1}
tasks = []

app = Flask(__name__)
# Disable flask printing
log = logging.getLogger('werkzeug')
log.disabled = True

@app.route("/", methods=['GET'])
def index():
    return app.send_static_file("index.html")

@app.route("/css/<name>", methods=['GET'])
def stylesheets(name):
    return app.send_static_file("./css/%s" % name)

@app.route("/assets/<name>", methods=['GET'])
def assets(name):
    return app.send_static_file("./assets/%s" % name)

@app.route('/channels', methods=['GET'])
@app.route('/channels/<name>', methods=['GET', 'POST'])
def channels(name=None):
    if request.method == 'GET':
        if type(name) == type(None):
            return jsonify(list(debug_log.keys()))
        else:
            return jsonify(debug_log[name] if name in debug_log else [])
    if request.method == 'POST':
        msg = request.json
        
        if name not in debug_log:
            debug_log[name] = []
        
        if type(msg) != type(None):
            debug_log[name].append(msg)

        return 'Debug Log Updated'

@app.route('/reset/<name>', methods=['POST'])
def reset(name):
    global debug_log

    if name.lower() == 'channels':
        debug_log = {channel: [] for channel in debug_log}

    return 'Reset Completed'

@app.route('/task', methods=['GET'])
@app.route('/task/<msgtype>', methods=['POST'])
def task(msgtype=None):
    global tasks

    if request.method == 'GET':
        return jsonify(tasks[0] if len(tasks) > 0 else MAIN_TASK)
    if request.method == 'POST':
        msg = request.json

        if msgtype.lower() == 'new':
            tasks.append({
                'task_name': msg['task_name'],
                'progress': 0, 
                'total_iters': float(msg['total_iters'])
            })

            return 'Task Added'

        if msgtype.lower() == 'update':
            tasks[0]['progress'] += float(msg['increment'])

            if tasks[0]['progress'] >= tasks[0]['total_iters']:
                tasks.pop(0)

            return 'Task Progressed'

@app.route('/timing/<channel>', methods=['POST'])
@app.route('/timing/<channel>', methods=['GET'])
def timing(channel=None):
    global timing_log

    if request.method == 'GET':
        return jsonify(timing_log[channel] if channel in timing_log else {})
    if request.method == 'POST':
        msg = request.json

        if channel not in timing_log:
            timing_log[channel] = {}
            
        channel_log = timing_log[channel] 
        
        fn_name = msg['function']
        if fn_name not in channel_log:
            channel_log[fn_name] = {
                'avg_duration': msg['duration'],
                'n_calls': 1
            }
        else:
            avg_duration = channel_log[fn_name]['avg_duration']
            n_calls = channel_log[fn_name]['n_calls']

            n = n_calls + 1
            mu = (msg['duration'] + n_calls * avg_duration) / n
            channel_log[fn_name]['avg_duration'] = mu
            channel_log[fn_name]['n_calls'] = n

        return 'call registered'
                
@app.route('/tests/<channel>', methods=['POST'])
@app.route('/tests/<channel>', methods=['GET'])
def tests(channel=None):
    global tests_log

    if request.method == 'GET':
        return jsonify(tests_log[channel] if channel in tests_log else {})

    if request.method == 'POST':
        msgs = request.json
        # msgs must be a list of test results
        for msg in msgs:
            if channel not in tests_log:
                tests_log[channel] = []

            channel_log = tests_log[channel] 

            if msg not in channel_log:
                channel_log.append(msg)

        return 'Tests logged'


if __name__ == "__main__":
    app.run(debug=True, host='0.0.0.0', port=8080)